//////////////////////////////////////////////////////////////////////////////
//
// AutoPitLot v2.0
// Obsuga magnetometru HMC5883 na magistrali I2C
//
// (c) Pit Lab
// http://www.pitlab.pl
//////////////////////////////////////////////////////////////////////////////
#include "LPC1768.h"
#include "sys_def.h"
#include "errors.h"
#include "config_fram.h"
#include "HMC5883.h"
#include <math.h>

// Obsuga magnetometru wymaga wykonania kilku czynnoci rozoonych w czasie
// 1) Wystartowanie konwersji trwajcej 6ms (slot czasowy 0)
// 2) Wystartowanie odczytu wykonanego pomiaru. Odczyt wykonuje si w tle w procedurze przerwania I2C i 
//    trwa ok. 4ms (slot 11)
// 3) Przepisania odczytanych w przerwaniu danych do zmiennych pomiarowych

//Czas od rozpoczcia pomiaru do gotowych danych 6ms

////////////////////////////////////////////////////////////////////////////////
// Budzi ukad i startuje konwersj magnetometru HMC5883
// Dane powinny pojawi si 6ms po starcie konwersji
// Parametry: 
// [i] chChan - kana magistrali I2C 0, 1 lub 2
// Zwraca: nic
// Czas zajcia magistrali I2C: 760us przy zegarze 38kHz
////////////////////////////////////////////////////////////////////////////////
unsigned char StartMagnetometerHMC(unsigned char chChan)
{
    unsigned char chErr;

    chData[0] = (1 << 0);   //Mode Select:0=Continuous-Measurement Mode, 1=Single-Measurement Mode, 2-3=Idle Mode.

    switch(chChan)
    {
    /*case I2C0: chErr = StartI2C0(0x3C, MODE, chData, 1, 0);   //rozpocznij pomiar na I2C0
        break;
    case I2C1: chErr = StartI2C1(0x3C, MODE, chData, 1, 0);   //rozpocznij pomiar na I2C1
        break;*/
    case I2C2: chErr = StartI2C2(0x3C, MODE, chData, 1, 0);   //rozpocznij pomiar na I2C2
        break;
    default: chErr = ERR_INVALID_DATA; break;
    }
    return chErr;
}


////////////////////////////////////////////////////////////////////////////////
// Inicjuje odczyt danych z magnetometru HMC5883
// Parametry: 
// [i] chChan - kana magistrali I2C 1 lub 2
// Zwraca: nic
// Czas zajcia magistrali I2C: 2,2ms przy zegarze 38kHz
////////////////////////////////////////////////////////////////////////////////
unsigned char ReadMagnetometerHMC(unsigned char chChan)
{
    unsigned char chErr;

    switch(chChan)
    {
    /*case I2C0: chErr = StartI2C0(0x3C, DATA_XH, chData, 0, 6);   //wylij polecenie zapisu wszystkich pomiarw
        break;
    case I2C1: chErr = StartI2C1(0x3C, DATA_XH, chData, 0, 6);   //wylij polecenie zapisu wszystkich pomiarw
        break;*/
    case I2C2: chErr = StartI2C2(0x3C, DATA_XH, chData, 0, 6);   //wylij polecenie odczytu wszystkich pomiarw
        break;
    default: chErr = ERR_INVALID_DATA; break;
    }
    return chErr;
}


////////////////////////////////////////////////////////////////////////////////
// Analizuje dane i liczy wartoci natenia pola magnetycznego we wszystkich 3 osiach
// Parametry: brak
// Zwraca: kod bdu
// Czas wykonywania 23us
////////////////////////////////////////////////////////////////////////////////
unsigned char CalculateMagnetometerHMC(void)
{
    volatile short sTemp;
    unsigned char x;

    if ((!chData[0] && !chData[1]) | (!chData[2] && !chData[3]) | (!chData[4] && !chData[5]))
        return ERR_INVALID_DATA;

    //odejmij teoretyczny offset zera i filtruj dolnoprzepustowo
    sTemp = 0x100 * chData[0] + chData[1];
    fVectMagnHMC[0] = sTemp;
    sTemp = 0x100 * chData[2] + chData[3];
    fVectMagnHMC[2] = sTemp;   
    sTemp = 0x100 * chData[4] + chData[5];
    fVectMagnHMC[1] = sTemp;

    //jeeli trwa pomiar korekcji magnetometru na magistrali I2C1 to zwr surowe dane
    if (sInitFlag & FIND_CORR2)
        return ERR_OK;

    for (x=0; x<3; x++)
        fVectMagnHMC[x] -= fCorrMagnHMC[x];

    //jeeli trwa kalibracja to uwzgldniaj tylko wspczynnik korekcyjny
    if (sInitFlag & CALIB_MAG2)
        return ERR_OK;

    //normalny pomiar z uwzglednieniem wszystkich wspczynikw korekcyjnych
    for (x=0; x<3; x++)
        fVectMagnHMC[x] = fVectMagnHMC[x] * fGainMagnHMC[x] - fOfstMagnHMC[x];     
    return ERR_OK;
}


////////////////////////////////////////////////////////////////////////////////
// Funkcja kapsukujca funkcj obliczeniow. Pozwala na uycie na zewntrz bez potrzeby publikowania danych
// Parametry: 
// [i] chChan - kana magistrali I2C 0, 1 lub 2
// Zwraca: kod bdu
////////////////////////////////////////////////////////////////////////////////
unsigned char MeasureMagnetometerHMC(unsigned char chChan)
{
     unsigned char x, chErr;

    switch(chChan)
    {
    case I2C2: 
        chErr = CalculateMagnetometerHMC();
        if (sInitFlag & CALIB_MAG2)     //jeeli trwa kalibracja to znajd wartoci maksymalne
        {
          for (x=0; x<3; x++)
          {
            if (fVectMagnHMC[x] > fMaxValMagn[x])
            {
              fMaxValMagn[x] = fVectMagnHMC[x];
              BeepMax();
            }
            if (fVectMagnHMC[x] < fMinValMagn[x])
            {
              fMinValMagn[x] = fVectMagnHMC[x];
              BeepMin();
            }
          }
        }
        break;

        default: 
            chErr = ERR_INVALID_DATA; 
            break;

    }
    return chErr;
}
////////////////////////////////////////////////////////////////////////////////
// Rozpoczyna i koczy kolejne etapy procedury kalibracyjnej magetometrw na obu magistralach
// Parametry: 
// [i] chEtap - etapy kalibracji: 
//  0 = zerowanie wspczynnikw korekcji na magistrali I2C1
//  1 = zerowanie wspczynnikw korekcji na magistrali I2C2
//  2 = ponowne wczytanie zewntrznie zapisanych zmiennych korekcji na magistrali I2C1
//  3 = ponowne wczytanie zewntrznie zapisanych zmiennych korekcji na magistrali I2C2
//  5 = ustawienie flagi kalibracji magnetometru na magistrali I2C1
//  6 = ustawienie flagi kalibracji magnetometru na magistrali I2C2
//  7 = kasowanie flagi kalibracji i obliczenie wspczynnikw kalibracji magnetometru na obu magistralach
// Zwraca: kod bdu
////////////////////////////////////////////////////////////////////////////////
unsigned char CalibrateMagnHMC(unsigned char chEtap)
{
    unsigned char x, chErr = ERR_OK;
    float fVect3D;
    
    switch (chEtap)
    {
    /*case 0: //zeruj korekcj na kanale I2C1
        sInitFlag |= FIND_CORR1;
        for (x=0; x<3; x++)
            fCorrMagnHMC[x] = 0.0;
        break;*/

    case 1: //zeruj korekcj na kanale I2C2
        sInitFlag |= FIND_CORR2;
        for (x=0; x<3; x++)
            fCorrMagnHMC[x] = 0.0;
        break;

   /* case 2:  //przeaduj konfiguracj  na kanale I2C1
        sInitFlag &= ~FIND_CORR1;
        for (x=0; x<3; x++)
        {
            chErr |= FramDataReadFloatValid(FAH_MAG_HMC_CORRX+x*4, &fCorrMagnHMC[x], VALM_CORR_MAGN, VALP_CORR_MAGN, VALD_CORR_MAGN, ERR_MAGN_CAL_VAL);
            chErr |= FramDataReadFloatValid(FAH_MAG_HMC_GAINX+x*4, &fGainMagnHMC[x], VALM_GAIN_MAGN, VALP_GAIN_MAGN, VALD_GAIN_MAGN, ERR_MAGN_CAL_VAL);           
            chErr |= FramDataReadFloatValid(FAH_MAG_HMC_OFSTX+x*4, &fOfstMagnHMC[x], VALM_OFST_MAGN, VALP_OFST_MAGN, VALD_OFST_MAGN, ERR_MAGN_CAL_VAL);            
        }
        break;*/

    case 3: //przeaduj konfiguracj  na kanale I2C2
        sInitFlag &= ~FIND_CORR2;
        for (x=0; x<3; x++)
        {
            chErr |= FramDataReadFloatValid(FAH_MAG_HMC_CORRX+x*4, &fCorrMagnHMC[x], VALM_CORR_MAGN, VALP_CORR_MAGN, VALD_CORR_MAGN, ERR_MAGN_CAL_VAL);
            chErr |= FramDataReadFloatValid(FAH_MAG_HMC_GAINX+x*4, &fGainMagnHMC[x], VALM_GAIN_MAGN, VALP_GAIN_MAGN, VALD_GAIN_MAGN, ERR_MAGN_CAL_VAL);
            chErr |= FramDataReadFloatValid(FAH_MAG_HMC_OFSTX+x*4, &fOfstMagnHMC[x], VALM_OFST_MAGN, VALP_OFST_MAGN, VALD_OFST_MAGN, ERR_MAGN_CAL_VAL);
        }
        break;

    /*case 5: //ustaw flag kalibracji magnetometru na magistrali I2C1
        sInitFlag |= CALIB_MAG1;
        for (x=0; x<3; x++)
        {
            fMinValMagn[x] = 0.0;
            fMaxValMagn[x] = 0.0;
        }
        break;*/

    case 6: //ustaw flag kalibracji magnetometru na magistrali I2C2
        sInitFlag |= CALIB_MAG2;
         for (x=0; x<3; x++)
        {
            fMinValMagn[x] = 0.0;
            fMaxValMagn[x] = 0.0;
        }
        break;

    case 7: //wykonaj kalibraj i kasuj flagi kalibracji magnetometru na magistralach I2C1 i I2C2
        /*if (sInitFlag & CALIB_MAG1)
        {
            //oblicz znormalizowan warto wektora 3D ekstremw pola magnetycznego
            fVect3D = sqrtf((fMaxValMagn[0] - fMinValMagn[0]) * (fMaxValMagn[0] - fMinValMagn[0])+
                            (fMaxValMagn[1] - fMinValMagn[1]) * (fMaxValMagn[1] - fMinValMagn[1])+
                            (fMaxValMagn[2] - fMinValMagn[2]) * (fMaxValMagn[2] - fMinValMagn[2]));

            //oblicz a nastpnie zapisz gain i ofset dla czujnika na magistrali I2C1
            for (x=0; x<3; x++)
            {
                fGainMagnHMC[x] = fVect3D / (fMaxValMagn[x] - fMinValMagn[x]);
                fOfstMagnHMC[x] = (fMaxValMagn[x] + fMinValMagn[x]) / 2 * fGainMagnHMC[x];
                FramDataWriteFloat(FAH_MAG_HMC_GAINX+x*4, fGainMagnHMC[x]);
                FramDataWriteFloat(FAH_MAG_HMC_OFSTX+x*4, fOfstMagnHMC[x]);
            }
            sInitFlag &= ~CALIB_MAG1;
        }*/
        if (sInitFlag & CALIB_MAG2)
        {
            //oblicz znormalizowan warto wektora 3D ekstremw pola magnetycznego
            fVect3D = sqrtf((fMaxValMagn[0] - fMinValMagn[0]) * (fMaxValMagn[0] - fMinValMagn[0])+
                            (fMaxValMagn[1] - fMinValMagn[1]) * (fMaxValMagn[1] - fMinValMagn[1])+
                            (fMaxValMagn[2] - fMinValMagn[2]) * (fMaxValMagn[2] - fMinValMagn[2]));

            //oblicz a nastpnie zapisz gain i offset dla czujnika na magistrali I2C2
            for (x=0; x<3; x++)
            {
                fGainMagnHMC[x] = fVect3D / (fMaxValMagn[x] - fMinValMagn[x]);
                fOfstMagnHMC[x] = ((fMaxValMagn[x] + fMinValMagn[x]) / 2) * fGainMagnHMC[x];
                FramDataWriteFloat(FAH_MAG_HMC_GAINX+x*4, fGainMagnHMC[x]);
                FramDataWriteFloat(FAH_MAG_HMC_OFSTX+x*4, fOfstMagnHMC[x]);
            }
            sInitFlag &= ~CALIB_MAG2;
        }
        break;

    case 8: //zakocz kalibracj i korekcj na wszystkich kanaach
        sInitFlag &= ~(FIND_CORR1 | FIND_CORR2 | CALIB_MAG1 | CALIB_MAG2);
        break;

    case 9: //inicjalizuj dane do wartoci domylnych na I2C1
    case 10: //inicjalizuj dane do wartoci domylnych na I2C2
        for (x=0; x<3; x++)
        {
            fGainMagnHMC[x] = (float)1.0;
            fOfstMagnHMC[x] = (float)0.0;
            FramDataWriteFloat(FAH_MAG_HMC_GAINX+x*4, fGainMagnHMC[x]);
            FramDataWriteFloat(FAH_MAG_HMC_OFSTX+x*4, fOfstMagnHMC[x]);
        }
        break;

    default: 
        chErr = ERR_INVALID_DATA;
    }
    return chErr;
}



////////////////////////////////////////////////////////////////////////////////
// Inicjuje zmienne konfiguracyjne
// Parametry: 
// Zwraca: kod bdu
////////////////////////////////////////////////////////////////////////////////
unsigned char InitMagnetometerHMC(void)
{
    unsigned char x, chErr = ERR_OK;   

    //sprawd obecno magnetometru na magistrali I2C2
    chErr = TestPresenceHMC5883(2);
    if (chErr == ERR_BRAK_MAGN)
        return chErr;

    //Ustaw tryb pracy na 30Hz
    //Configuration Register A
    chData[0] = (0 << 0)|   //Measurement Configuration: 0=Normal measurement configuration, 1=Positive bias configuration, 2=Negative bias configuration
                (5 << 2)|   //Data Output Rate:6=75Hz, 5=30, 4=15, 3=7,5, 2=3 
                (2 << 5)|   //Select number of samples averaged per measurement output: 00 = 1; 01 = 2; 10 = 4; 11 = 8
                (0 << 7);   //This bit must be cleared for correct operation.
    //Configuration Register B
    chData[1] = (1 << 5);   //Gain Configuration: 0=0,88 Gaussa, 1=1,3; 2=1,9; , 3=2,5; 4=4; 5=4,7; 6=5,6; 7=8,1 Gaussa
    //Mode Register
    chData[2] = (0 << 0)|   //Mode Select:0=Continuous-Measurement Mode, 1=Single-Measurement Mode, 2-3=Idle Mode.
                (0 << 2);   //Bits 2-7 must be cleared for correct operation.
    chErr = StartI2C2(0x3C, CONF_A, chData, 3, 0);  //zapisz 3 rejestry jedn transmisj
    while (!chI2C2End);

    //odczytaj wartoci kalibracyjne magnetometru
    for (x=0; x<3; x++)
    {
        //offset zera pocztkowego magnetometru
        chErr |= FramDataReadFloatValid((unsigned short)(FAH_MAG_HMC_CORRX+x*4), &fCorrMagnHMC[x], VALM_CORR_MAGN, VALP_CORR_MAGN, VALD_CORR_MAGN, ERR_MAGN_CAL_VAL);

        //wzmocnienie wskaza magnetometru
        chErr |= FramDataReadFloatValid((unsigned short)FAH_MAG_HMC_GAINX+x*4, &fGainMagnHMC[x], VALM_GAIN_MAGN, VALP_GAIN_MAGN, VALD_GAIN_MAGN, ERR_MAGN_CAL_VAL);
        
        //offset kompensacji hard/soft iron magnetometru
        chErr |= FramDataReadFloatValid((unsigned short)FAH_MAG_HMC_OFSTX+x*4, &fOfstMagnHMC[x], VALM_OFST_MAGN, VALP_OFST_MAGN, VALD_OFST_MAGN, ERR_MAGN_CAL_VAL);
    }

    if (!chErr)
        sInitFlag |= IDONE_MAG; //zakonczono kalibracj magnetometru
    return chErr;
}


////////////////////////////////////////////////////////////////////////////////
// Testuje obecno magnetometru HMC5883 na magistrali korzystajc z Identification Register A,B,C
// Parametry: 
// [i] chChan - kana magistrali I2C 0, 1 lub 2
// Zwraca: nic
// Czas zajcia magistrali I2C: ms przy zegarze 50kHz
////////////////////////////////////////////////////////////////////////////////
unsigned char TestPresenceHMC5883(unsigned char chChan)
{
    unsigned char chErr;
    volatile int nStartTime = T0TC;  //inicjuj pomiar czasu

    switch(chChan)
    {
    case 0: //I2C0
        chErr = StartI2C0(0x3C, ID_A, chData, 0, 3); 
        while (!chI2C0End)
            if (T0TC > nStartTime + SERVO_CNT_FREQ/(2*FREQ_SLOT_CZAS))   //czekaj maksymalnie 1/2 slotu czasowego
                return ERR_BRAK_MAGN;
        break;
    case 1: //I2C1
        chErr = StartI2C1(0x3C, ID_A, chData, 0, 3);
        while (!chI2C1End)
            if (T0TC > nStartTime + SERVO_CNT_FREQ/(2*FREQ_SLOT_CZAS))   //czekaj maksymalnie 1/2 slotu czasowego
                return ERR_BRAK_MAGN;
        break;
    case 2: //I2C2
        chErr = StartI2C2(0x3C, ID_A, chData, 0, 3);
        while (!chI2C2End)
            if (T0TC > nStartTime + SERVO_CNT_FREQ/(2*FREQ_SLOT_CZAS))   //czekaj maksymalnie 1/2 slotu czasowego
                return ERR_BRAK_MAGN;
        break;
    default: chErr = ERR_INVALID_DATA; break;
    }
    if ((chData[0] != 'H') && (chData[1] != '4') && (chData[2] != '3'))
        chErr = ERR_BRAK_MAGN;

    return chErr;
}
